mod benchmarks;
mod stress;
mod text;

use clap::{Args, Parser, Subcommand};

#[derive(Parser)]
#[clap(name = "Bit Goblin Benchmark", author, version, about = "Bit Goblin's hardware benchmarking tool.", long_about = None)]
#[clap(propagate_version = true)]
struct Cli {
    #[clap(subcommand)]
    command: Commands,

    #[clap(short = 'l', long, default_value_t = 1, help = "Number of times to run test. Default = 1", global = true)]
    loopcount: u8,
}

#[derive(Subcommand)]
enum Commands {
    // CPU benchmarks subcommand
    #[clap(name = "cpu", about = "CPU benchmarks and stress tests.")]
    Cpu(Cpu),
    // memory benchmarks subcommand
    #[clap(name = "memory", about = "Memory benchmarks and stress tests.")]
    Mem(Mem),
    // disk benchmarks subcommand
    #[clap(name = "disk", about = "Hard drive and SSD benchmarks.")]
    Disk(Disk),
    // games benchmarks subcommand
    #[clap(name = "games", about = "Benchmark your system with games.")]
    Games(Games),
    // network benchmarks subcommand
    #[clap(name = "network", about = "Test various aspects of your network.")]
    Net(Net),
}

#[derive(Args)]
struct Cpu {
    #[clap(subcommand)]
    cpu_commands: CpuCommands,
}

#[derive(Subcommand)]
enum CpuCommands {
    // CPU stress test subcommand
    #[clap(name = "stress", about = "Stress test the CPU with math!")]
    StressTest {
        #[clap(short = 'r', long, default_value_t = 5, help = "Length of time (in minutes) to run the stress test. Defaults to 5")]
        runtime: u16,
        #[clap(short = 't', long, default_value_t = 0, help = "Number of threads to use; defaults to CPU's max thread count. Defaults to 0 (automatic)")]
        threads: usize,
    },

    // CPU benchmark test subcommand
    #[clap(name = "benchmark", about = "Benchmark the CPU")]
    Benchmark {
        #[clap(short = 'm', long, default_value_t = 4)]
        threads: u8,
        #[clap(short = 'o', long, default_value_t = 10000)]
        maxprime: u32,
    },
}

#[derive(Args)]
struct Mem {
    #[clap(subcommand)]
    mem_commands: MemCommands,
}

#[derive(Subcommand)]
enum MemCommands {
    // Memory benchmark test subcommand
    #[clap(name = "benchmark", about = "Benchmark the system memory")]
    Benchmark {
        #[clap(short = 'm', long, default_value_t = String::from("seq"))]
        accessmode: String,
        #[clap(short = 'o', long, default_value_t = String::from("write"))]
        operation: String,
        #[clap(short = 'b', long, default_value_t = String::from("1K"))]
        blocksize: String,
        #[clap(short = 's', long, default_value_t = String::from("100G"))]
        totalsize: String,
    },
}

#[derive(Args)]
struct Disk {
    #[clap(subcommand)]
    disk_commands: DiskCommands,
}

#[derive(Subcommand)]
enum DiskCommands {
    // sequential disk read subcommand
    #[clap(name = "read_seq", about = "Sequential disk read speed test.")]
    ReadSeqTest {
        #[clap(short = 't', long, default_value_t = String::from("/tmp/disk-test.tmp"))]
        tempfile: String,
        #[clap(short = 's', long, default_value_t = 15)]
        size: u8,
    },

    // random disk read subcommand
    #[clap(name = "read_rand", about = "Random 4K disk read speed test.")]
    ReadRandTest {
        #[clap(short = 't', long, default_value_t = String::from("/tmp/disk-test.tmp"))]
        tempfile: String,
        #[clap(short = 's', long, default_value_t = 15)]
        size: u8,
    },

    // sequential disk write subcommand
    #[clap(name = "write_seq", about = "Write a large file to determine sequential disk write speeds.")]
    WriteSeqTest {
        #[clap(short = 't', long, default_value_t = String::from("/tmp/disk-test.tmp"))]
        tempfile: String,
        #[clap(short = 's', long, default_value_t = 15)]
        size: u8,
    },

    // random 4K disk write subcommand
    #[clap(name = "write_rand", about = "Write a bunch of smallfiles to determine random disk write speeds.")]
    WriteRandTest {
        #[clap(short = 't', long, default_value_t = String::from("/tmp/disk-test.tmp"))]
        tempfile: String,
        #[clap(short = 's', long, default_value_t = 15)]
        size: u8,
    },
}

#[derive(Parser)]
struct Games {
    #[structopt(subcommand)]
    games_commands: GamesCommands,
}

#[derive(Subcommand)]
enum GamesCommands {
    // Civilization 6 AI benchmark subcommand
    #[clap(name = "civ6_ai", about = "Run the Civilization 6 AI benchmark via Steam.")]
    Civ6AI {},

    // CS:GO benchmark subcommand
    #[clap(name = "csgo", about = "Run the CS:GO game benchmark via Steam.")]
    CSGO {},

    // Deus Ex: Mankind Divided benchmark subcommand
    #[clap(name = "demd", about = "Run the Deus Ex: Mankind Divided game benchmark via Steam.")]
    DEMD {},
}

#[derive(Parser)]
struct Net {
    #[structopt(subcommand)]
    net_commands: NetCommands,
}

#[derive(Subcommand)]
enum NetCommands {
    // bandwidth test subcommand
    #[clap(name = "bandwidth", about = "Uses iperf to test network bandwidth.")]
    Bandwidth {
        #[clap(short = 'a', long, required = true)]
        host: String,
    },

    // jitter subcommand
    #[clap(name = "jitter", about = "Ping a host to determine network jitter.")]
    Jitter {
        #[clap(short = 'a', long, default_value_t = String::from("8.8.8.8"))]
        address: String,
        #[clap(short = 'c', long, default_value_t = 100)]
        count: u16,
        #[clap(short = 'i', long, default_value_t = 1000)]
        interval: u16,
    },

    // latency subcommand
    #[clap(name = "latency", about = "Ping a host to determine network latency.")]
    Latency {
        #[clap(short = 'a', long, default_value_t = String::from("8.8.8.8"))]
        address: String,
        #[clap(short = 'c', long, default_value_t = 100)]
        count: u16,
        #[clap(short = 'i', long, default_value_t = 1000)]
        interval: u16,
    },
}

fn main() {
    let cli = Cli::parse();

    // map subcommands back to the main command
    match &cli.command {
        Commands::Cpu(args) => match &args.cpu_commands {
            CpuCommands::StressTest { runtime, threads } => stress::cpu::cpu_stress_math(*runtime, *threads),
            CpuCommands::Benchmark { threads, maxprime } => benchmarks::cpu::sysbench(threads, maxprime),
        },

        Commands::Mem(args) => match &args.mem_commands {
            MemCommands::Benchmark { accessmode, operation, blocksize, totalsize } => benchmarks::memory::sysbench(accessmode, operation, blocksize, totalsize),
        },

        Commands::Disk(args) => match &args.disk_commands {
            DiskCommands::ReadSeqTest { tempfile, size } => {
                for i in 0..cli.loopcount {
                    println!("Test run number {}.", i + 1);
                    benchmarks::disk::disk_read_seq_test(tempfile, size);
                }
            },
            DiskCommands::ReadRandTest { tempfile, size } => {
                for i in 0..cli.loopcount {
                    println!("Test run number {}.", i + 1);
                    benchmarks::disk::disk_read_rand_test(tempfile, size);
                }
            },
            DiskCommands::WriteSeqTest { tempfile, size } => {
                for i in 0..cli.loopcount {
                    println!("Test run number {}.", i + 1);
                    benchmarks::disk::disk_write_seq_test(tempfile, size);
                }
            },
            DiskCommands::WriteRandTest { tempfile, size } => {
                for i in 0..cli.loopcount {
                    println!("Test run number {}.", i + 1);
                    benchmarks::disk::disk_write_rand_test(tempfile, size);
                }
            },
        },

        Commands::Games(args) => match &args.games_commands {
            GamesCommands::Civ6AI {} => benchmarks::games::run_civ6_ai_benchmark(),
            GamesCommands::CSGO {} => benchmarks::games::run_csgo_benchmark(),
            GamesCommands::DEMD {} => benchmarks::games::run_demd_benchmark(),
        },

        Commands::Net(args) => match &args.net_commands {
            NetCommands::Bandwidth { host } => {
                for i in 0..cli.loopcount {
                    println!("Test run number {}.", i + 1);
                    benchmarks::network::bandwidth_test(host);
                }
            },
            NetCommands::Jitter { address, count, interval } => {
                for i in 0..cli.loopcount {
                    println!("Test run number {}.", i + 1);
                    benchmarks::network::jitter_test(address, count, interval);
                }
            },
            NetCommands::Latency { address, count, interval } => {
                for i in 0..cli.loopcount {
                    println!("Test run number {}.", i + 1);
                    benchmarks::network::latency_test(address, count, interval);
                }
            },
        },
    }
}
